from prometheus_client import Counter, Gauge, Summary, Histogram # noqa


HTTP_SERVER_REQUESTS_TOTAL = Counter(
    'http_server_requests_total',
    'Total requests',
    ['method', 'host', 'path', 'code']
)

HTTP_SERVER_EXCEPTIONS_TOTAL = Counter(
    'http_server_exceptions_total',
    'Total number of exceptions',
    ['error']
)

HTTP_SERVER_REQUEST_DURATION_SECONDS = Histogram(
    'http_server_request_duration_seconds', 'Server request duration in seconds',
    ['method', 'host', 'path', 'code']
)



import os
import pytest
import time
from flask import Flask

from gds_metrics import GDSMetrics


FAKE_APP_ID = "123"
SLOW_REQUEST_DURATION = 0.1


@pytest.fixture
def app(mocker):
    flask_app = Flask(__name__)
    ctx = flask_app.app_context()
    ctx.push()

    mocker.patch.dict(
        os.environ,
        {'VCAP_APPLICATION': '{"application_id": "' + FAKE_APP_ID + '"}'}
    )
    mocker.patch('prometheus_client.generate_latest', return_value='Prometheus metrics response')

    @flask_app.route('/')
    def index():
        return 'index'

    @flask_app.route('/slow_request')
    def slow_request():
        time.sleep(SLOW_REQUEST_DURATION)
        return 'slow request'

    @flask_app.route('/exception')
    def exception():
        raise Exception()

    yield flask_app

    ctx.pop()


@pytest.fixture(scope='function')
def client(app):
    with app.test_request_context(), app.test_client() as client:
        metrics = GDSMetrics()
        metrics.init_app(app)

        yield client


@pytest.fixture(scope='function')
def client_without_basic_auth(app, mocker):
    with app.test_request_context(), app.test_client() as client:
        mocker.patch.dict(
            os.environ,
            {
                'VCAP_APPLICATION': '{"application_id": "' + FAKE_APP_ID + '"}',
                'METRICS_BASIC_AUTH': 'false'
            }
        )

        metrics = GDSMetrics()
        metrics.init_app(app)

        yield client


@pytest.fixture(scope='function')
def client_without_env_app_id(app, mocker):
    with app.test_request_context(), app.test_client() as client:
        mocker.patch.dict(os.environ, {'VCAP_APPLICATION': '{}'})

        metrics = GDSMetrics()
        metrics.init_app(app)

        yield client



import gzip
import hmac
import json
import os
from time import monotonic

from flask import abort, g, request, Response
from flask.signals import got_request_exception, request_finished

# set multiprocess temp directory before we import prometheus_client
os.environ.setdefault('prometheus_multiproc_dir', '/tmp') # noqa

import prometheus_client
from prometheus_client import multiprocess, CollectorRegistry

from .metrics import ( # noqa proxy metric types imports
    Counter,
    Gauge,
    Summary,
    Histogram,
    HTTP_SERVER_EXCEPTIONS_TOTAL,
    HTTP_SERVER_REQUEST_DURATION_SECONDS,
    HTTP_SERVER_REQUESTS_TOTAL,
)


class GDSMetrics(object):

    def __init__(self):
        self.metrics_path = os.environ.get('PROMETHEUS_METRICS_PATH', '/metrics')
        if os.environ.get("METRICS_BASIC_AUTH", "true") == "true":
            self.auth_token = json.loads(os.environ.get("VCAP_APPLICATION", "{}")).get("application_id")
        else:
            self.auth_token = False

        self.registry = CollectorRegistry()
        multiprocess.MultiProcessCollector(self.registry)

    def init_app(self, app):
        app.add_url_rule(self.metrics_path, 'metrics', self.metrics_endpoint)

        app.before_request(self.before_request)
        request_finished.connect(self.teardown_request, sender=app)
        got_request_exception.connect(self.handle_exception, sender=app)

    def metrics_endpoint(self):
        if self.auth_token:
            auth_header = request.headers.get('Authorization', '')
            if not auth_header:
                abort(401)
            elif not hmac.compare_digest(auth_header, 'Bearer {}'.format(self.auth_token)):
                abort(403)

        response = Response(
            prometheus_client.generate_latest(self.registry),
            mimetype='text/plain; version=0.0.4; charset=utf-8',
            headers={
                'Cache-Control': 'no-cache, no-store, max-age=0, must-revalidate',
            }
        )

        accept_encoding = request.headers.get('Accept-Encoding', '')

        if 'gzip' not in accept_encoding.lower():
            return response

        response.data = gzip.compress(response.data)
        response.headers['Content-Encoding'] = 'gzip'
        response.headers['Content-Length'] = len(response.data)
        response.headers['Vary'] = 'Accept-Encoding'

        return response

    def before_request(self, *args, **kwargs):
        g._gds_metrics_start_time = monotonic()

    def teardown_request(self, sender, response, *args, **kwargs):
        if hasattr(g, "_gds_metrics_start_time"):
            # There is not a guarantee that `g._gds_metrics_start_time` will exist
            # as our `before_request` function will not run if the flask app has
            # another `before_request` that runs before ours and throws an exception
            resp_time = monotonic() - g._gds_metrics_start_time
            HTTP_SERVER_REQUEST_DURATION_SECONDS.labels(
                request.method,
                request.host,
                request.url_rule.rule if request.url_rule else 'No endpoint',
                response.status_code
            ).observe(resp_time)

        HTTP_SERVER_REQUESTS_TOTAL.labels(
            request.method,
            request.host,
            request.url_rule.rule if request.url_rule else 'No endpoint',
            response.status_code
        ).inc()

        return response

    def handle_exception(self, sender, exception, *args, **kwargs):
        HTTP_SERVER_EXCEPTIONS_TOTAL.labels(type(exception)).inc()



from unittest.mock import Mock

from gds_metrics import GDSMetrics

from flask import Flask, g


def test_teardown_requests_doesnt_record_duration_metric_if_gds_metrics_start_time_not_set(mocker):
    app = Flask(__name__)
    with app.test_request_context():
        assert not hasattr(g, "_gds_metrics_start_time")
        duration_metric = mocker.patch('gds_metrics.HTTP_SERVER_REQUEST_DURATION_SECONDS.labels')
        count_metric = mocker.patch('gds_metrics.HTTP_SERVER_REQUESTS_TOTAL.labels')
        gds_metrics = GDSMetrics()

        gds_metrics.teardown_request(sender=Mock(), response=Mock())

        assert not duration_metric.called
        assert count_metric.called



import pytest
from tests.conftest import FAKE_APP_ID, SLOW_REQUEST_DURATION

VALID_AUTH_HEADER = 'Authorization', 'Bearer {}'.format(FAKE_APP_ID)


@pytest.mark.parametrize('auth_header,expected_status', [
    (None, 401),
    ([VALID_AUTH_HEADER], 200),
    ([('Authorization', 'Bearer invalid-token')], 403),
])
def test_auth_header_returns_expected_response(client, auth_header, expected_status, mocker):
    response = client.get(
        '/metrics',
        headers=auth_header
    )
    assert response.status_code == expected_status


def test_no_basic_auth_env_flag_returns_200(client_without_basic_auth):
    response = client_without_basic_auth.get('/metrics')
    assert response.status_code == 200


def test_metrics_path_without_app_id_env_does_not_need_auth(client_without_env_app_id):
    response = client_without_env_app_id.get('/metrics')
    assert response.status_code == 200


def test_app_path_does_not_need_auth(client):
    response = client.get('/')
    assert response.status_code == 200


def test_compresses_response_with_gzip_header(client):
    response = client.get(
        '/metrics',
        headers=[('Accept-Encoding', 'gzip'), (VALID_AUTH_HEADER)]
    )
    assert response.headers['Content-Encoding'] == 'gzip'


def test_does_not_compress_response_without_gzip_header(client):
    response = client.get(
        '/metrics',
        headers=[VALID_AUTH_HEADER]
    )
    assert response.headers.get('Content-Encoding') is None


def test_gzip_request_header_does_not_affect_other_paths(app, client):
    response = client.get(
        '/',
        headers=[('Accept-Encoding', 'gzip'), (VALID_AUTH_HEADER)]
    )
    assert response.headers.get('Content-Encoding') is None


def test_calls_histogram_observe(client, mocker):
    histogram_observe = mocker.patch(
        'gds_metrics.metrics.HTTP_SERVER_REQUEST_DURATION_SECONDS._wrappedClass.observe')

    client.get(
        '/slow_request',
        headers=[(VALID_AUTH_HEADER)]
    )
    assert histogram_observe.called
    # cannot define exact duration so expect at least the specified duration for the slow_request
    assert histogram_observe.call_args_list[0][0][0] > SLOW_REQUEST_DURATION and \
        histogram_observe.call_args_list[0][0][0] < SLOW_REQUEST_DURATION + 0.05


def test_requests_increases_request_counter(client, mocker):
    counter_inc = mocker.patch(
        'gds_metrics.metrics.HTTP_SERVER_REQUESTS_TOTAL._wrappedClass.inc')
    client.get(
        '/',
        headers=[(VALID_AUTH_HEADER)]
    )
    assert counter_inc.called


def test_exception_increases_exception_counter(client, mocker):
    counter_inc = mocker.patch(
        'gds_metrics.metrics.HTTP_SERVER_EXCEPTIONS_TOTAL._wrappedClass.inc')

    response = client.get(
        '/exception',
        headers=[(VALID_AUTH_HEADER)]
    )
    assert response.status_code == 500
    assert counter_inc.called



"""
Library for Prometheus instrumentation in Python/Flask based apps
"""

import re
import ast
from setuptools import setup, find_packages


_version_re = re.compile(r'__version__\s+=\s+(.*)')

with open('gds_metrics/version.py', 'rb') as f:
    version = str(ast.literal_eval(_version_re.search(
        f.read().decode('utf-8')).group(1)))

setup(
    name='gds-metrics',
    version=version,
    url='https://github.com/alphagov/gds_metrics_python',
    license='MIT',
    author='Government Digital Service',
    description='Library for Prometheus instrumentation in Python/Flask based apps',
    long_description=__doc__,
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "prometheus_client>=0.2.0",
        "Flask>=0.10",
        "blinker>=1.4",
    ],
    # for running pytest as `python setup.py test`, see
    # http://doc.pytest.org/en/latest/goodpractices.html#integrating-with-setuptools-python-setup-py-test-pytest-runner
    setup_requires=[
        'pytest-runner'
    ],
    test_suite="tests",
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: System Administrators",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Topic :: System :: Monitoring",
        "License :: OSI Approved :: MIT License",
    ],
    keywords='gds reliability engineering metrics',
)
